"use strict";
var __assign = (this && this.__assign) || function () {
    __assign = Object.assign || function(t) {
        for (var s, i = 1, n = arguments.length; i < n; i++) {
            s = arguments[i];
            for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p))
                t[p] = s[p];
        }
        return t;
    };
    return __assign.apply(this, arguments);
};
var __rest = (this && this.__rest) || function (s, e) {
    var t = {};
    for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p) && e.indexOf(p) < 0)
        t[p] = s[p];
    if (s != null && typeof Object.getOwnPropertySymbols === "function")
        for (var i = 0, p = Object.getOwnPropertySymbols(s); i < p.length; i++) {
            if (e.indexOf(p[i]) < 0 && Object.prototype.propertyIsEnumerable.call(s, p[i]))
                t[p[i]] = s[p[i]];
        }
    return t;
};
var __spreadArrays = (this && this.__spreadArrays) || function () {
    for (var s = 0, i = 0, il = arguments.length; i < il; i++) s += arguments[i].length;
    for (var r = Array(s), k = 0, i = 0; i < il; i++)
        for (var a = arguments[i], j = 0, jl = a.length; j < jl; j++, k++)
            r[k] = a[j];
    return r;
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
var interfaces_1 = require("./interfaces");
var get_1 = __importDefault(require("lodash/get"));
var type_formatter_1 = require("./type-formatter");
var assign_1 = __importDefault(require("lodash/assign"));
var defaults_1 = __importDefault(require("lodash/defaults"));
var mapKeys_1 = __importDefault(require("lodash/mapKeys"));
var pick_1 = __importDefault(require("lodash/pick"));
var keys_1 = __importDefault(require("lodash/keys"));
var cloneDeep_1 = __importDefault(require("lodash/cloneDeep"));
var isFunction_1 = __importDefault(require("lodash/isFunction"));
var forEach_1 = __importDefault(require("lodash/forEach"));
var set_1 = __importDefault(require("lodash/set"));
/**
 * Maps from a Sails route path of the form `/path/:id` to a
 * Swagger path of the form `/path/{id}`. and the path variables
 */
exports.generateSwaggerPath = function (path) {
    var variables = [];
    var swaggerPath = path
        .split('/')
        .map(function (v) {
        // TODO: update this to accept optional route param e.g id?
        // simply remove the ? in [^/:?]
        var match = v.match(/^:([^/:?]+)\??$/);
        if (match) {
            variables.push(match[1]);
            return "{" + match[1] + "}";
        }
        return v;
    })
        .join('/');
    return { path: swaggerPath, variables: variables };
};
/**
 * Generate Swagger schema content describing the specified Sails attribute.
 * TODO: add test to this function
 *
 * @see https://swagger.io/docs/specification/data-models/
 * @param {Record<string, any>} attribute Sails model attribute specification as per `Model.js` file
 */
// eslint-disable-next-line @typescript-eslint/no-explicit-any
exports.generateAttributeSchema = function (attribute) {
    var ai = attribute || {}, sts = type_formatter_1.swaggerTypes;
    var type = attribute.type || 'string';
    var columnType = get_1.default(attribute, ['autoMigrations', 'columnType']);
    var autoIncrement = get_1.default(attribute, ['autoMigrations', 'autoIncrement']);
    var schema = {};
    if (ai.model) {
        assign_1.default(schema, {
            description: "JSON dictionary representing the **" + ai.model + "** instance or FK when creating / updating / not populated",
            // '$ref': '#/components/schemas/' + ai.model,
            oneOf: [
                { '$ref': '#/components/schemas/' + ai.model },
            ],
        });
    }
    else if (ai.collection) {
        assign_1.default(schema, {
            description: "Array of **" + ai.collection + "**'s or array of FK's when creating / updating / not populated",
            type: 'array',
            items: { '$ref': '#/components/schemas/' + ai.collection },
        });
    }
    else if (type == 'number') {
        var t = autoIncrement ? sts.integer : sts.double;
        if (ai.isInteger) {
            t = sts.integer;
        }
        else if (columnType) {
            var ct = columnType;
            if (ct.match(/int/i))
                t = sts.integer;
            else if (ct.match(/long/i))
                t = sts.long;
            else if (ct.match(/float/i))
                t = sts.float;
            else if (ct.match(/double/i))
                t = sts.double;
            else if (ct.match(/decimal/i))
                t = sts.double;
        }
        assign_1.default(schema, t);
    }
    else if (type == 'boolean') {
        assign_1.default(schema, sts.boolean);
    }
    else if (type == 'json') {
        assign_1.default(schema, sts.string);
    }
    else if (type == 'ref') {
        var t = sts.string;
        if (columnType) {
            var ct = columnType;
            if (ct.match(/timestamp/i))
                t = sts.datetime;
            else if (ct.match(/datetime/i))
                t = sts.datetime;
            else if (ct.match(/date/i))
                t = sts.date;
        }
        assign_1.default(schema, t);
    }
    else { // includes =='string'
        assign_1.default(schema, sts.string);
    }
    var isIP = false;
    if (schema.type == 'string') {
        var v = ai.validations;
        if (v) {
            if (v.isEmail)
                schema.format = 'email';
            if (v.isIP) {
                isIP = true;
                schema.format = 'ipv4';
            }
            if (v.isURL)
                schema.format = 'uri';
            if (v.isUUID)
                schema.format = 'uuid';
        }
    }
    var annotations = [];
    // annotate format with Sails autoCreatedAt/autoUpdatedAt
    if (schema.type == 'string' && schema.format == 'date-time') {
        if (ai.autoCreatedAt)
            annotations.push('autoCreatedAt');
        else if (ai.autoUpdatedAt)
            annotations.push('autoUpdatedAt');
    }
    // process Sails --> Swagger attribute mappings as per sailAttributePropertiesMap
    defaults_1.default(schema, mapKeys_1.default(pick_1.default(ai, keys_1.default(type_formatter_1.sailAttributePropertiesMap)), function (v, k) { return type_formatter_1.sailAttributePropertiesMap[k]; }));
    // process Sails --> Swagger attribute mappings as per validationsMap
    defaults_1.default(schema, mapKeys_1.default(pick_1.default(ai.validations, keys_1.default(type_formatter_1.validationsMap)), function (v, k) { return get_1.default(type_formatter_1.validationsMap, "" + k); }));
    // copy default into example if present
    if (schema.default && !schema.example) {
        schema.example = schema.default;
    }
    // process final autoMigrations: autoIncrement, unique
    if (autoIncrement)
        annotations.push('autoIncrement');
    if (get_1.default(attribute, ['autoMigrations', 'unique'])) {
        schema.uniqueItems = true;
    }
    // represent Sails `isIP` as one of ipv4/ipv6
    if (schema.type == 'string' && isIP) {
        schema = {
            description: 'ipv4 or ipv6 address',
            oneOf: [
                cloneDeep_1.default(schema),
                assign_1.default(cloneDeep_1.default(schema), { format: 'ipv6' }),
            ]
        };
    }
    if (annotations.length > 0) {
        var s = "Note Sails special attributes: " + annotations.join(', ');
        schema.description = schema.description ? "\n\n" + s : s;
    }
    // note: required --> required[] (not here, needs to be done at model level)
    delete schema.common_name;
    return schema;
};
/**
 * Generate Swagger schema content describing specified Sails models.
 * @see https://swagger.io/docs/specification/data-models/
 * @param models parsed Sails models as per `parsers.parseModels()`
 * @returns
 */
exports.generateSchemas = function (models) {
    return Object.keys(models)
        .reduce(function (schemas, identiy) {
        var model = models[identiy];
        var schema = {
            description: get_1.default(model, 'swagger.description', "Sails ORM Model **" + model.globalId + "**"),
            required: [],
        };
        var attributes = model.attributes || {};
        schema.properties = Object.keys(attributes).reduce(function (props, attributeName) {
            var attribute = model.attributes[attributeName];
            props[attributeName] = exports.generateAttributeSchema(attribute);
            if (attribute.required)
                schema.required.push(attributeName);
            return props;
        }, {});
        if (schema.required.length <= 0)
            delete schema.required;
        schemas[model.identity] = schema;
        return schemas;
    }, {});
};
/**
 * Generate Swagger schema content describing specified Sails routes/actions.
 *
 * @see https://swagger.io/docs/specification/paths-and-operations/
 *
 * TODO: break down this function into smaller methods and add tests separately
 *
 * @param routes
 * @param templates
 * @param defaultsValues
 * @param action2s
 * @param specification
 */
exports.generatePaths = function (routes, templates, defaultsValues, action2s, specification) {
    var paths = {};
    var tags = specification.tags;
    var components = specification.components;
    if (!components.parameters) {
        components.parameters = {};
    }
    var addTags = function (toAdd) {
        var newTags = toAdd.filter(function (newTag) { return tags.find(function (tag) { return newTag.name === tag.name; }); });
        tags.push.apply(tags, newTags);
    };
    var mergeComponents = function (toMerge) {
        for (var key in toMerge) {
            var componentName = key;
            if (!components[componentName]) {
                components[componentName] = {};
            }
            defaults_1.default(components[componentName], toMerge[componentName]);
        }
    };
    var _loop_1 = function (route) {
        var path = route.path;
        var pathEntry;
        if (route.middlewareType === interfaces_1.MiddlewareType.BLUEPRINT && route.model) {
            var template_1 = templates[route.action] || {};
            var subst_1 = function (str) { return str ? str.replace('{globalId}', route.model.globalId) : ''; };
            // handle special case of PK parameter
            var parameters = __spreadArrays((template_1.parameters || [])).map(function (parameter) {
                if (parameter === 'primaryKeyPathParameter') {
                    var primaryKey = route.model.primaryKey;
                    var attributeInfo = route.model.attributes[primaryKey];
                    var pname = 'ModelPKParam-' + route.model.identity;
                    if (components.parameters && !components.parameters[pname]) {
                        components.parameters[pname] = {
                            in: 'path',
                            name: primaryKey,
                            required: true,
                            schema: exports.generateAttributeSchema(attributeInfo),
                            description: subst_1('The desired **{globalId}** record\'s primary key value'),
                        };
                    }
                    return { $ref: '#/components/parameters/' + pname };
                }
                return parameter;
            });
            pathEntry = __assign({ summary: subst_1(template_1.summary), description: subst_1(template_1.description), externalDocs: template_1.externalDocs || undefined, tags: get_1.default(route, 'swagger.tags', []), parameters: parameters, responses: cloneDeep_1.default(defaultsValues.responses || {}) }, (route.swagger || {}));
            var isParam_1 = function (inType, name) { return !!pathEntry.parameters.find(function (parameter) { return parameter.in == inType && parameter.name == name; }); };
            var modifiers_1 = {
                addPopulateQueryParam: function () {
                    if (isParam_1('query', 'populate') || route.associations.length == 0)
                        return;
                    pathEntry.parameters.push({
                        in: 'query',
                        name: 'populate',
                        required: false,
                        schema: {
                            type: 'string',
                            example: __spreadArrays(['false'], (route.associations.map(function (row) { return row.alias; }) || [])).join(','),
                        },
                        description: 'If specified, overide the default automatic population process.'
                            + ' Accepts a comma-separated list of attribute names for which to populate record values,'
                            + ' or specify `false` to have no attributes populated.',
                    });
                },
                addSelectQueryParam: function () {
                    if (isParam_1('query', 'select'))
                        return;
                    pathEntry.parameters.push({
                        in: 'query',
                        name: 'select',
                        required: false,
                        schema: {
                            type: 'string',
                            example: __spreadArrays(keys_1.default(route.model.attributes || {})).join(','),
                        },
                        description: 'The attributes to include in the result, specified as a comma-delimited list.'
                            + ' By default, all attributes are selected.'
                            + ' Not valid for plural (“collection”) association attributes.',
                    });
                },
                addOmitQueryParam: function () {
                    if (isParam_1('query', 'omit'))
                        return;
                    pathEntry.parameters.push({
                        in: 'query',
                        name: 'omit',
                        required: false,
                        schema: {
                            type: 'string',
                            example: __spreadArrays(keys_1.default(route.model.attributes || {})).join(','),
                        },
                        description: 'The attributes to exclude from the result, specified as a comma-delimited list.'
                            + ' Cannot be used in conjuction with `select`.'
                            + ' Not valid for plural (“collection”) association attributes.',
                    });
                },
                addModelBodyParam: function () {
                    if (pathEntry.requestBody)
                        return;
                    pathEntry.requestBody = {
                        description: subst_1('JSON dictionary representing the {globalId} instance to create.'),
                        required: true,
                        content: {
                            'application/json': {
                                schema: { "$ref": "#/components/schemas/" + route.model.identity }
                            },
                        },
                    };
                },
                addResultOfArrayOfModels: function () {
                    assign_1.default(pathEntry.responses['200'], {
                        description: subst_1(template_1.resultDescription),
                        content: {
                            'application/json': {
                                schema: {
                                    type: 'array',
                                    items: { '$ref': '#/components/schemas/' + route.model.identity },
                                },
                            },
                        },
                    });
                },
                addAssociationPathParam: function () {
                    if (isParam_1('path', 'association'))
                        return;
                    pathEntry.parameters.splice(1, 0, {
                        in: 'path',
                        name: 'association',
                        required: true,
                        schema: {
                            type: 'string',
                            enum: route.aliases,
                        },
                        description: 'The name of the association',
                    });
                },
                addAssociationFKPathParam: function () {
                    if (isParam_1('path', 'fk'))
                        return;
                    pathEntry.parameters.push({
                        in: 'path',
                        name: 'fk',
                        required: true,
                        schema: {
                            oneOf: route.associationsPrimaryKeyAttribute.map(function (ai) { return exports.generateAttributeSchema(ai); }),
                        },
                        description: 'The desired target association record\'s foreign key value'
                    });
                },
                addAssociationResultOfArray: function () {
                    var models = route.aliases.map(function (a) {
                        var assoc = route.associations.find(function (_assoc) { return _assoc.alias == a; });
                        return assoc ? (assoc.collection || assoc.model) : a;
                    });
                    assign_1.default(pathEntry.responses['200'], {
                        description: subst_1(template_1.resultDescription),
                        content: {
                            'application/json': {
                                schema: {
                                    type: 'array',
                                    // items: { type: 'any' },
                                    items: {
                                        oneOf: models.map(function (model) {
                                            return { '$ref': '#/components/schemas/' + model };
                                        }),
                                    },
                                },
                            },
                        },
                    });
                },
                addResultOfModel: function () {
                    assign_1.default(pathEntry.responses['200'], {
                        description: subst_1(template_1.resultDescription),
                        content: {
                            'application/json': {
                                schema: { '$ref': '#/components/schemas/' + route.model.identity },
                            },
                        },
                    });
                },
                addResultNotFound: function () {
                    assign_1.default(pathEntry.responses['404'], {
                        description: subst_1(template_1.notFoundDescription || 'Not found'),
                    });
                },
                addResultValidationError: function () {
                    assign_1.default(pathEntry.responses['400'], {
                        description: subst_1('Validation errors; details in JSON response'),
                    });
                },
                addFksBodyParam: function () {
                    if (pathEntry.requestBody)
                        return;
                    pathEntry.requestBody = {
                        description: 'The primary key values (usually IDs) of the child records to use as the new members of this collection',
                        required: true,
                        content: {
                            'application/json': {
                                schema: {
                                    type: 'array',
                                    items: {
                                        oneOf: route.associationsPrimaryKeyAttribute.map(function (ai) { return exports.generateAttributeSchema(ai); }),
                                    }
                                },
                            },
                        },
                    };
                },
            };
            // apply changes for blueprint action
            (template_1.modifiers || []).map(function (modifier) {
                if (isFunction_1.default(modifier))
                    modifier(template_1, route, pathEntry, tags, components); // custom modifier
                else
                    modifiers_1[modifier](); // standard modifier
            });
            assign_1.default(pathEntry.responses['500'], {
                description: 'Internal server error',
            });
        }
        else {
            pathEntry = __assign({ parameters: [], responses: cloneDeep_1.default(defaultsValues.responses || {}) }, (route.swagger || {}));
            var action2 = action2s[route.action];
            if (action2) {
                var patternVariables_1 = route.variables || [];
                if (action2.inputs) {
                    forEach_1.default(action2.inputs, function (value, key) {
                        var _in = patternVariables_1.indexOf(key) >= 0 ? 'path' : 'query';
                        var isExisting = pathEntry.parameters.find(function (p) { return p.in === _in && p.name === key; });
                        if (isExisting) {
                            return;
                        }
                        var description = value.description, attribute = __rest(value, ["description"]);
                        pathEntry.parameters.push({
                            in: _in,
                            name: key,
                            required: value.required || false,
                            schema: exports.generateAttributeSchema(attribute),
                            description: description
                        });
                    });
                }
                if (action2.exits) {
                    var exitResponses_1 = {};
                    forEach_1.default(action2.exits, function (exit, exitName) {
                        var hasDefaults = type_formatter_1.actions2Responses[exitName];
                        if (!hasDefaults) {
                            return;
                        }
                        var _a = exit.statusCode, statusCode = _a === void 0 ? hasDefaults.statusCode : _a, _b = exit.description, description = _b === void 0 ? hasDefaults.description : _b, headers = exit.headers, content = exit.content, links = exit.links;
                        exitResponses_1[statusCode] = {
                            description: description,
                            content: content,
                            headers: headers,
                            links: links
                        };
                    });
                    pathEntry.responses = __assign(__assign({}, exitResponses_1), pathEntry.responses);
                    forEach_1.default(pathEntry.responses, function (v, k) {
                        if (!v.description)
                            v.description = get_1.default(exitResponses_1, [k, 'description'], '-');
                    });
                }
            }
            // actions2 summary and description
            defaults_1.default(pathEntry, {
                summary: get_1.default(route, ['actions2', 'friendlyName']) || route.path || '',
                description: get_1.default(route, ['actions2', 'description']) || undefined,
            });
        }
        if (route.variables) {
            // first resolve '$ref' parameters
            var resolved_1 = pathEntry.parameters.map(function (parameter) {
                var ref = parameter['$ref'];
                if (!ref)
                    return parameter;
                var _prefix = '#/components/parameters/';
                if (ref.startsWith(_prefix)) {
                    ref = ref.substring(_prefix.length);
                    var dereferenced = components.parameters[ref];
                    return dereferenced ? dereferenced : parameter;
                }
            });
            // now add patternVariables that don't already exist
            route.variables.map(function (v) {
                var existing = resolved_1.find(function (p) { return p.in == 'path' && p.name == v; });
                if (existing)
                    return;
                pathEntry.parameters.push({
                    in: 'path',
                    name: v,
                    required: true,
                    schema: { type: 'string' },
                    description: "Route pattern variable `" + v + "`",
                });
            });
        }
        addTags(get_1.default(route.swagger, 'tags', []));
        mergeComponents(get_1.default(route.swagger, 'components', {}));
        set_1.default(paths, [path, route.verb], pathEntry);
    };
    for (var _i = 0, routes_1 = routes; _i < routes_1.length; _i++) {
        var route = routes_1[_i];
        _loop_1(route);
    }
    return paths;
};
