"use strict";
var __assign = (this && this.__assign) || function () {
    __assign = Object.assign || function(t) {
        for (var s, i = 1, n = arguments.length; i < n; i++) {
            s = arguments[i];
            for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p))
                t[p] = s[p];
        }
        return t;
    };
    return __assign.apply(this, arguments);
};
var __spreadArrays = (this && this.__spreadArrays) || function () {
    for (var s = 0, i = 0, il = arguments.length; i < il; i++) s += arguments[i].length;
    for (var r = Array(s), k = 0, i = 0; i < il; i++)
        for (var a = arguments[i], j = 0, jl = a.length; j < jl; j++, k++)
            r[k] = a[j];
    return r;
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
var interfaces_1 = require("../interfaces");
var swagger_jsdoc_1 = __importDefault(require("swagger-jsdoc"));
var cloneDeep_1 = __importDefault(require("lodash/cloneDeep"));
var uniqBy_1 = __importDefault(require("lodash/uniqBy"));
var mapKeys_1 = __importDefault(require("lodash/mapKeys"));
var get_1 = __importDefault(require("lodash/get"));
var uniq_1 = __importDefault(require("lodash/uniq"));
var defaults_1 = __importDefault(require("lodash/defaults"));
var forEach_1 = __importDefault(require("lodash/forEach"));
var uniqBy = uniqBy_1.default;
var componentReference = [
    "schemas",
    "parameters",
    "securitySchemes",
    "requestBodies",
    "responses",
    "headers",
    "examples",
    "links",
    "callbacks"
];
exports.bluerprintActions = ['findone', 'find', 'create', 'update', 'destroy', 'populate', 'add', 'remove', 'replace'];
exports.mergeSwaggerSpec = function (destination, source) {
    var dest = cloneDeep_1.default(destination);
    dest.tags = uniqBy(__spreadArrays((get_1.default(source, 'tags', [])), (get_1.default(dest, 'tags', []))), 'name');
    dest.components = componentReference.reduce(function (mergedComponents, ref) {
        var type = get_1.default(source, "components." + ref);
        if (type) {
            mergedComponents[ref] = __assign(__assign({}, (get_1.default(dest.components, ref, {}))), type);
        }
        return mergedComponents;
    }, dest.components || {});
    return dest;
};
exports.mergeSwaggerPaths = function (dest, sourcePaths) {
    // strip off leading '/' from both source and destination swagger action, swagger-doc returns paths with leading /path-name: {}
    var swaggerActions = mapKeys_1.default(cloneDeep_1.default(dest || {}), function (_, key) { return key.replace(/^\//, ''); });
    var paths = mapKeys_1.default(cloneDeep_1.default(sourcePaths), function (_, key) { return key.replace(/^\//, ''); });
    // unique action keys
    var actions = uniq_1.default(Object.keys(swaggerActions).concat(Object.keys(paths)));
    return actions.reduce(function (merged, key) {
        var destination = swaggerActions[key];
        var source = paths[key];
        merged[key] = __assign(__assign({}, (destination || {})), (source || {}));
        return merged;
    }, {});
};
exports.loadSwaggerDocComments = function (filePath) {
    return new Promise(function (resolve, reject) {
        try {
            var opts = {
                definition: {
                    openapi: '3.0.0',
                    info: { title: 'dummy', version: '0.0.0' },
                },
                apis: [filePath],
            };
            var specification = swagger_jsdoc_1.default(opts);
            resolve(specification);
        }
        catch (err) {
            reject(err);
        }
    });
};
exports.removeViewRoutes = function (routes) {
    return Object.keys(routes).reduce(function (cleanedRoutes, routeAddress) {
        var target = routes[routeAddress];
        var isView = !!get_1.default(target, 'view');
        if (!isView) {
            cleanedRoutes[routeAddress] = target;
        }
        return cleanedRoutes;
    }, {});
};
exports.getBlueprintAllowedMiddlewareRoutes = function (routes) {
    return routes.filter(function (route) {
        var middlewareType = get_1.default(route, 'options._middlewareType', '');
        var type = middlewareType.split(':')[0];
        return type === interfaces_1.MiddlewareType.BLUEPRINT;
    });
};
exports.normalizeRouteControllerName = function (name) {
    if (name && !name.endsWith('Controller')) {
        return "" + name.charAt(0).toUpperCase() + name.substring(1) + "Controller";
    }
    return name;
};
// see: https://sailsjs.com/documentation/concepts/routes/custom-routes#?controller-action-target-syntax
exports.normalizeRouteControllerTarget = function (target) {
    if (typeof target === 'string') {
        // for foo.myGoAction and FooController.myGoAction
        var dotParams = target.split('.');
        if (dotParams.length > 1) {
            var controller = exports.normalizeRouteControllerName(dotParams[0]);
            var action = dotParams[1];
            return { controller: controller, action: action };
        }
        // foo/go-action
        var slashParams = target.split('/');
        if (slashParams.length > 1) {
            return { action: target };
        }
    }
    else {
        return {
            controller: exports.normalizeRouteControllerName(target.controller),
            action: target.action,
            swagger: target.swagger
        };
    }
    throw new TypeError("Unable to normalize route: " + target + ", see: https://sailsjs.com/documentation/concepts/routes/custom-routes for allowed routes pattern");
};
exports.getSwaggerAction = function (object, action) {
    return get_1.default(object, "swagger.actions." + action);
};
exports.getAction2Paths = function (routes) {
    return routes
        .filter(function (route) {
        return !route.controller;
    })
        .map(function (route) {
        return route.action;
    });
};
exports.getUniqueTagsFromPath = function (paths) {
    var referencedTags = new Set();
    for (var path in paths) {
        var pathDefinition = paths[path];
        for (var verb in pathDefinition) {
            var verbDefinition = pathDefinition[verb];
            if (verbDefinition.tags) {
                verbDefinition.tags.forEach(function (tag) { return referencedTags.add(tag); });
            }
        }
    }
    return referencedTags;
};
exports.mergeComponents = function (existingComponents, models, controllers, action2s) {
    var components = cloneDeep_1.default(existingComponents);
    var merge = function (toMerge) {
        for (var key in toMerge) {
            var componentName = key;
            if (!components[componentName]) {
                components[componentName] = toMerge[componentName];
            }
            defaults_1.default(components[componentName], toMerge[componentName]);
        }
    };
    forEach_1.default(models, function (model) {
        merge(get_1.default(model, 'swagger.components', {}));
    });
    forEach_1.default(controllers, function (controller) {
        merge(get_1.default(controller, 'swagger.components', {}));
    });
    forEach_1.default(action2s, function (action2) {
        merge(get_1.default(action2, 'swagger.components', {}));
    });
    return components;
};
exports.mergeTags = function (existingTags, models, controllers, action2s) {
    var tags = cloneDeep_1.default(existingTags);
    var addTags = function (toAdd) {
        var newTags = toAdd.filter(function (newTag) { return !tags.find(function (tag) { return newTag.name === tag.name; }); });
        tags.push.apply(tags, newTags);
    };
    forEach_1.default(models, function (model) {
        addTags(get_1.default(model, 'swagger.tags', []));
    });
    forEach_1.default(controllers, function (controller) {
        addTags(get_1.default(controller, 'swagger.tags', []));
    });
    forEach_1.default(action2s, function (action2) {
        addTags(get_1.default(action2, 'swagger.tags', []));
    });
    return tags;
};
exports.getActionNameFromPath = function (actionPath) {
    return actionPath.split('/').pop();
};
