"use strict";
var __assign = (this && this.__assign) || function () {
    __assign = Object.assign || function(t) {
        for (var s, i = 1, n = arguments.length; i < n; i++) {
            s = arguments[i];
            for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p))
                t[p] = s[p];
        }
        return t;
    };
    return __assign.apply(this, arguments);
};
Object.defineProperty(exports, "__esModule", { value: true });
var generators_1 = require("../lib/generators");
var chai_1 = require("chai");
// eslint-disable-next-line @typescript-eslint/no-var-requires
var userModel = require('../../api/models/User');
describe('Generators', function () {
    describe('generateSwaggerPath', function () {
        it('should return swagger open api version path', function (done) {
            var params = generators_1.generateSwaggerPath('/pets/:id');
            chai_1.expect(params.path).to.equal('/pets/{id}');
            done();
        });
        it('should return path params as variables', function (done) {
            var params = generators_1.generateSwaggerPath('/pets/:id/:anotherID');
            chai_1.expect(params.variables).to.deep.equal(['id', 'anotherID']);
            done();
        });
    });
    describe('generateSchemas', function () {
        // TODO: attribute with relationship with other models
        it('should generate OpenAPIV3 accepted schema type', function (done) {
            var actual = generators_1.generateSchemas({ user: __assign(__assign({}, userModel), { identity: 'user' }) });
            var expected = {
                user: {
                    description: "Sails ORM Model **undefined**",
                    required: ['names'],
                    properties: {
                        id: {
                            type: "number",
                            format: "double"
                        },
                        names: {
                            type: "string",
                            example: "First Middle Last"
                        },
                        email: {
                            type: "string",
                            description: "Just any old email"
                        },
                        sex: {
                            type: "string"
                        },
                        ageLimit: {
                            type: "number",
                            format: "double"
                        }
                    }
                }
            };
            chai_1.expect(actual).to.deep.equal(expected);
            done();
        });
    });
});
