"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __generator = (this && this.__generator) || function (thisArg, body) {
    var _ = { label: 0, sent: function() { if (t[0] & 1) throw t[1]; return t[1]; }, trys: [], ops: [] }, f, y, t, g;
    return g = { next: verb(0), "throw": verb(1), "return": verb(2) }, typeof Symbol === "function" && (g[Symbol.iterator] = function() { return this; }), g;
    function verb(n) { return function (v) { return step([n, v]); }; }
    function step(op) {
        if (f) throw new TypeError("Generator is already executing.");
        while (_) try {
            if (f = 1, y && (t = op[0] & 2 ? y["return"] : op[0] ? y["throw"] || ((t = y["return"]) && t.call(y), 0) : y.next) && !(t = t.call(y, op[1])).done) return t;
            if (y = 0, t) op = [op[0] & 2, t.value];
            switch (op[0]) {
                case 0: case 1: t = op; break;
                case 4: _.label++; return { value: op[1], done: false };
                case 5: _.label++; y = op[1]; op = [0]; continue;
                case 7: op = _.ops.pop(); _.trys.pop(); continue;
                default:
                    if (!(t = _.trys, t = t.length > 0 && t[t.length - 1]) && (op[0] === 6 || op[0] === 2)) { _ = 0; continue; }
                    if (op[0] === 3 && (!t || (op[1] > t[0] && op[1] < t[3]))) { _.label = op[1]; break; }
                    if (op[0] === 6 && _.label < t[1]) { _.label = t[1]; t = op; break; }
                    if (t && _.label < t[2]) { _.label = t[2]; _.ops.push(op); break; }
                    if (t[2]) _.ops.pop();
                    _.trys.pop(); continue;
            }
            op = body.call(thisArg, _);
        } catch (e) { op = [6, e]; y = 0; } finally { f = t = 0; }
        if (op[0] & 5) throw op[1]; return { value: op[0] ? op[1] : void 0, done: true };
    }
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
var path_1 = __importDefault(require("path"));
var chai_1 = require("chai");
var utils_1 = require("../lib/utils");
var sailsConfig = {
    paths: {
        models: '../../api/models'
    }
};
describe('Utils', function () {
    describe('mergeSwaggerDoc', function () {
        var destJSON = "{\n            \"tags\": [\n              {\n                \"name\": \"Users\",\n                \"description\": \"dest description\"\n              }\n            ],\n            \"components\": {\n              \"schemas\": {\n                \"Users\": {\n                  \"properties\": {}\n                },\n                \"Clients\": {}\n              }\n            }\n          }";
        var sourceJSON = "{\n            \"tags\": [\n              {\n                \"name\": \"Clients\"\n              },\n              {\n                \"name\": \"Users\",\n                \"description\": \"some fancy description\"\n              }\n            ],\n            \"components\": {\n              \"schemas\": {\n                \"Users\": {}\n              }\n            }\n          }";
        var dest = JSON.parse(destJSON);
        var source = JSON.parse(sourceJSON);
        var actual = utils_1.mergeSwaggerSpec(dest, source);
        it('should not mutate destination and source swagger attribute', function (done) {
            chai_1.expect(dest).to.deep.equal(JSON.parse(destJSON));
            chai_1.expect(source).to.deep.equal(JSON.parse(sourceJSON));
            done();
        });
        it(' should merge tags and source tag properties should always override destination if a tag is existing', function (done) {
            chai_1.expect(actual.tags).to.deep.equal(source.tags);
            done();
        });
        it('should merge components and source component properties should override component that is existing', function (done) {
            chai_1.expect(actual.components, 'destination Users schema with properties is updated to take dest {}').to.deep.equal({
                schemas: {
                    Users: {},
                    Clients: {}
                }
            });
            done();
        });
    });
    describe('mergeSwaggerPaths', function () {
        var dest = {
            find: {
                description: 'something about description',
                summary: 'summary about find'
            }
        };
        var source = {
            '/find': {
                description: 'something with /'
            }
        };
        var merged = utils_1.mergeSwaggerPaths(dest, source);
        it('should remove only leading / from merged paths', function (done) {
            chai_1.expect(Object.keys(merged).every(function (key) { return key.charAt(0) !== '/'; })).to.be.true;
            done();
        });
        it('should merge and take /find and find action as the same', function (done) {
            chai_1.expect(Object.keys(merged), 'contain only one key after merging').to.deep.equal(['find']);
            chai_1.expect(merged.find, 'contain both source and destination keys').to.have.all.keys('summary', 'description');
            chai_1.expect(merged.find.description, 'source should override destination params').to.equal(source['/find'].description);
            done();
        });
    });
    describe('loadSwaggerDocComments', function () { return __awaiter(void 0, void 0, void 0, function () {
        var fixturePath, doc;
        return __generator(this, function (_a) {
            switch (_a.label) {
                case 0:
                    fixturePath = require.resolve(path_1.default.join(sailsConfig.paths.models, 'User'));
                    return [4 /*yield*/, utils_1.loadSwaggerDocComments(fixturePath)];
                case 1:
                    doc = _a.sent();
                    it('should generate openapi v3.0.0 spec doc from swagger-doc comment', function (done) {
                        chai_1.expect(doc.openapi).equal('3.0.0');
                        done();
                    });
                    return [2 /*return*/];
            }
        });
    }); });
    describe('removeViewRoutes', function () {
        it('should remove route targets with views', function (done) {
            chai_1.expect(utils_1.removeViewRoutes({
                'path/:id': 'address.foo',
                'get users/:id': { action: 'index' },
                'get users/me': { view: 'brochure/about' }
            })).to.deep.equal({
                'path/:id': 'address.foo',
                'get users/:id': { action: 'index' },
            });
            done();
        });
    });
    describe('normalizeRouteControllerName', function () {
        it('should normalize undefined route controller name to undefined', function (done) {
            chai_1.expect(utils_1.normalizeRouteControllerName(undefined)).to.undefined;
            done();
        });
        it('should normalize route controller name from foo to FooController', function (done) {
            chai_1.expect(utils_1.normalizeRouteControllerName('foo')).to.equal('FooController');
            done();
        });
        it('should normalize route controller name from FooController to FooController', function (done) {
            chai_1.expect(utils_1.normalizeRouteControllerName('FooController')).to.equal('FooController');
            done();
        });
    });
    describe('normalizeRouteControllerTarget', function () {
        it('should normalize FooController.myGoAction', function (done) {
            chai_1.expect(utils_1.normalizeRouteControllerTarget('FooController.myGoAction')).to.deep.equal({ controller: 'FooController', action: 'myGoAction' });
            done();
        });
        it('should normalize foo.myGoAction', function (done) {
            chai_1.expect(utils_1.normalizeRouteControllerTarget('foo.myGoAction')).to.deep.equal({ controller: 'FooController', action: 'myGoAction' });
            done();
        });
        it('should normalize foo/go-action', function (done) {
            chai_1.expect(utils_1.normalizeRouteControllerTarget('foo/go-action')).to.deep.equal({ action: 'foo/go-action' });
            done();
        });
        it("should normalize {action: 'index'}", function (done) {
            chai_1.expect(utils_1.normalizeRouteControllerTarget({ action: 'index' })).to.deep.equal({ action: 'index', controller: undefined, swagger: undefined });
            done();
        });
        it("should normalize {action: 'foo/go-action'}", function (done) {
            chai_1.expect(utils_1.normalizeRouteControllerTarget({ action: 'foo/go-action' })).to.deep.equal({ action: 'foo/go-action', controller: undefined, swagger: undefined });
            done();
        });
        it('should throw type error if route target is not among the accepted pattern', function (done) {
            chai_1.expect(function () { return utils_1.normalizeRouteControllerTarget('something-bind'); }).throw();
            done();
        });
    });
    describe('getSwaggerAction', function () {
        it('should not throw if model or controller is undefined', function (done) {
            // eslint-disable-next-line @typescript-eslint/no-explicit-any
            var object = undefined;
            chai_1.expect(utils_1.getSwaggerAction(object, 'example/action')).to.equal(undefined);
            done();
        });
        it('should return action if it exist in either model or controller', function (done) {
            var action = { ke: 'string' };
            // eslint-disable-next-line @typescript-eslint/no-explicit-any
            var object = { swagger: { actions: { action: action } } };
            chai_1.expect(utils_1.getSwaggerAction(object, 'action')).to.deep.equal(action);
            done();
        });
    });
    var models = {
        user: {
            swagger: {
                components: {
                    schemas: { modelSchema: {} }
                },
                tags: [{ name: 'modelTag' }]
            }
        }
    };
    var controllers = {
        UserController: {
            swagger: {
                components: {
                    schemas: { controllerSchema: {} }
                },
                tags: [{ name: 'controllerTag' }]
            }
        }
    };
    var action2 = {
        'action/action2': {
            swagger: {
                components: {
                    schemas: { action2Schema: {} }
                },
                tags: [{ name: 'action2Tag' }]
            }
        }
    };
    describe('mergeComponents', function () {
        it("should merge models,controllers and action2 swagger.components if it's existing", function (done) {
            var components = { schemas: { existingSchema: {} } };
            var actual = utils_1.mergeComponents(components, models, controllers, action2);
            chai_1.expect(Object.keys(actual.schemas)).to.deep.equal(['existingSchema', 'modelSchema', 'controllerSchema', 'action2Schema'], 'merge components props to existing components');
            done();
        });
    });
    describe('mergeTags', function () {
        it("should merge models,controllers and action2 swagger.tags if it's existing", function (done) {
            var tags = [{ name: 'existing' }];
            var actual = utils_1.mergeTags(tags, models, controllers, action2);
            chai_1.expect(actual).to.deep.equal([tags[0], models.user.swagger.tags[0], controllers.UserController.swagger.tags[0], action2['action/action2'].swagger.tags[0]], 'merge tags props to existing tags');
            done();
        });
    });
    describe('getActionNameFromPath', function () {
        it('should get action2 filename from action2path', function (done) {
            chai_1.expect(utils_1.getActionNameFromPath('subdir/actions2')).to.be.equal('actions2');
            done();
        });
    });
});
